// Released under the MIT licence.
// See LICENCE.txt for details.

#include "Backend.h"
#include "../../Misc.h"

#include <csignal>
#include <stdint.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdbool.h>
#include <sys/audioio.h>
#include <pthread.h>

static const int parkSignal = SIGUSR1;
static pthread_t mainThread;
static int sndfp;
static void (*parent_callback)(long *stream, size_t frames_total);
static bool audioDone = false;

static void* soundcheckthread(void*) {
    unsigned long for_as_long_as_it_seems_we_should = 0;
    uint16_t out[0x400 * 2];
    while(1) {
        bool ready = false;
        audio_prinfo_t audioInfo;
        ioctl(sndfp, AUDIO_GETINFO, &audioInfo);
        int oldeof = audioInfo.eof;

        usleep(for_as_long_as_it_seems_we_should);

        while(ready == false) {
            ioctl(sndfp, AUDIO_GETINFO, &audioInfo);
            if(audioInfo.eof > oldeof) {
                ready = true;
            }
        }

        long mix[0x400 * 2];
        memset(mix, 0, 0x400 * 2);

        // park the main thread since the game probably expects it to not be
        // doing anything else while we're requesting audio frames.
        audioDone = false;
        pthread_kill(mainThread, parkSignal);

        parent_callback(mix, 0x400 * 2);

        // unpark the main thread now that we're not using game state anymore
        audioDone = true;

        write(sndfp, mix, 0x400 * 2);

        for_as_long_as_it_seems_we_should = (0x400 / 48000) * 1000000;
    }
}

static void threadPark(int signo) {
    while(audioDone == false);
}

unsigned long SoftwareMixerBackend_Init(void (*callback)(long *stream, size_t frames_total)) {
    sndfp = open("/dev/audio", 0, O_WRONLY);

    if(sndfp == -1) {
        Backend_PrintError("Failed to open audio device.\n");
        return 0;
    }

    audio_prinfo_t audioInfo;
    ioctl(sndfp, AUDIO_GETINFO, &audioInfo);
    audioInfo.sample_rate = 48000;
    audioInfo.channels = 2;
    audioInfo.precision = 16;
    audioInfo.encoding = AUDIO_ENCODING_LINEAR;
    audioInfo.port = AUDIO_SPEAKER;
    ioctl(sndfp, AUDIO_SETINFO, &audioInfo);

    if(signal(parkSignal, threadPark) == SIG_ERR) {
        Backend_PrintError("Failed to register thread park signal handler for audio system.\n");
        return 0;
    }

    mainThread = pthread_self();

    pthread_t thread;
    pthread_attr_t thread_attr;
    pthread_attr_init(&thread_attr);

    pthread_create(&thread, &thread_attr, soundcheckthread, (void*)NULL);

    return 48000; // return the sampling rate
}

void SoftwareMixerBackend_Deinit(void) {
    close(sndfp);
}

bool SoftwareMixerBackend_Start(void) {
    return true;
}

void SoftwareMixerBackend_LockMixerMutex(void) {
}

void SoftwareMixerBackend_UnlockMixerMutex(void) {
}

void SoftwareMixerBackend_LockOrganyaMutex(void) {
}

void SoftwareMixerBackend_UnlockOrganyaMutex(void) {
}
